from cacahuate.models import User


class BaseUser:

    def get_user(self):
        identifier = self.get_identifier()

        # fetchs redis mirror user if there is None then creates one
        user = User.get_by('identifier', identifier)

        if user is None:
            user = User(identifier=identifier).save()

        return user

    def get_identifier(self):
        raise NotImplementedError('Must be implemented in subclasses')

    def get_human_name(self):
        raise NotImplementedError('Must be implemented in subclasses')

    def get_x_info(self, notification_backend):
        raise NotImplementedError('Must be implemented in subclasses')


class BaseAuthProvider:

    def __init__(self, config):
        self.config = config

    def authenticate(self, **credentials) -> BaseUser:
        raise NotImplementedError('Must be implemented in subclasses')


class BaseHierarchyProvider:

    def __init__(self, config):
        self.config = config

    def validate_user(self, user, **params):
        ''' given a user, should rise an exception if the user does not match
        the hierarchy conditions required via params '''
        raise NotImplementedError('Must be implemented in subclasses')

    def find_users(self, **params) -> [BaseUser]:
        ''' given the params, retrieves the
        user identifiers that match them '''
        raise NotImplementedError('Must be implemented in subclasses')



from typing import Iterator, TextIO, Callable
import os
from xml.dom import pulldom
from xml.dom.minidom import Element

from .errors import ProcessNotFound, ElementNotFound, MalformedProcess
from .mark import comment

XML_ATTRIBUTES = {
    'public': lambda a: a == 'true',
    'author': str,
    'date': str,
    'name': str,
    'description': lambda x: x,
    'start-node': lambda x: x,
}


class Xml:

    def __init__(self, config, filename):
        try:
            self.id, self.version, _ = filename.split('.')
        except ValueError:
            raise MalformedProcess(
                'Name of process is invalid, must be name.version.xml'
            )

        self.versions = [self.version]
        self.filename = filename
        self.config = config
        self.parser = pulldom.parse(
            open(os.path.join(config['XML_PATH'], filename))
        )

        try:
            info_node = next(self)
        except StopIteration:
            raise MalformedProcess('This process lacks the process-info node')

        if info_node.tagName != 'process-info':
            raise MalformedProcess('process-info node must be the first node')

        for attr, func in XML_ATTRIBUTES.items():
            try:
                node = info_node.getElementsByTagName(attr)[0]
            except IndexError:
                raise MalformedProcess(
                    'Process\' metadata lacks node {}'.format(attr)
                )

            if node.firstChild is not None:
                node.normalize()
                setattr(self, attr, func(node.firstChild.nodeValue))
            else:
                # the text child of <element></element> is the empty string
                setattr(self, attr, func(''))

    @classmethod
    def load(cls, config: dict, common_name: str, direct=False) -> TextIO:
        ''' Loads an xml file and returns the corresponding TextIOWrapper for
        further usage. The file might contain multiple versions so the latest
        one is chosen.

        common_name is the prefix of the file to find. If multiple files with
        the same prefix are found the last in lexicographical order is
        returned.'''
        if direct:
            # skip looking for the most recent version
            return Xml(config, common_name)

        files = reversed(sorted(os.listdir(config['XML_PATH'])))

        for filename in files:
            if filename.startswith(common_name):
                return Xml(config, filename)
        else:
            raise ProcessNotFound(common_name)

    def __next__(self):
        ''' Returns an inerator over the nodes and edges of a process defined
        by the xmlfile descriptor. Uses XMLPullParser so no memory is consumed
        for this task. '''

        ITERABLES = ('node', 'connector', 'process-info')

        for event, node in self.parser:
            if event == pulldom.START_ELEMENT and node.tagName in ITERABLES:
                self.parser.expandNode(node)

                return node

        raise StopIteration

    def __iter__(self):
        return self

    def find(self, testfunc: Callable[[Element], bool]) -> Element:
        ''' Given an interator returned by the previous function, tries to find
        the first node matching the given condition '''
        for element in self:
            if testfunc(element):
                return element

        raise ElementNotFound(
            'node or edge matching the given condition was not found'
        )

    def start_node(self) -> Element:
        ''' Returns the starting node '''
        start_node_id = getattr(self, 'start-node')
        start_node = self.find(lambda e: e.getAttribute('id') == start_node_id)

        return start_node

    @classmethod
    def list(cls, config):
        # Get all processes
        files = reversed(sorted(os.listdir(config['XML_PATH'])))

        # Load only the oldest processes
        processes = []

        for filename in files:
            try:
                id, version, _ = filename.split('.')
            except ValueError:
                continue

            try:
                xml = cls.load(config, filename, direct=True)
            except ProcessNotFound:
                continue
            except MalformedProcess:
                continue

            if not xml.public:
                continue

            if len(processes) == 0 or processes[-1].id != id:
                processes.append(xml)
            else:
                processes[-1].versions.append(version)

        return processes

    def to_json(self):
        return {
            'id': self.id,
            'version': self.version,
            'author': self.author,
            'date': self.date,
            'name': self.name,
            'description': self.description,
            'versions': self.versions,
        }


def get_ref(el: Element):
    if el.getAttribute('id'):
        return '#' + el.getAttribute('id')
    elif el.getAttribute('class'):
        return '.' + el.getAttribute('class')

    return None


def resolve_params(filter_node, execution=None):
    computed_params = {}

    for param in filter_node.getElementsByTagName('param'):
        if execution is not None and param.getAttribute('type') == 'ref':
            user_ref = param.firstChild.nodeValue.split('#')[1].strip()

            try:
                actor = next(
                    execution.proxy.actors.q().filter(ref='#'+user_ref)
                )

                value = actor.proxy.user.get().identifier
            except StopIteration:
                value = None
        else:
            value = param.firstChild.nodeValue

        computed_params[param.getAttribute('name')] = value

    return computed_params


@comment
def etree_from_list(root: Element, nodes: [Element]) -> 'ElementTree':
    ''' Returns a built ElementTree from the list of its members '''
    root = Element(root.tag, attrib=root.attrib)
    root.extend(nodes)

    return ElementTree(root)


@comment
def nodes_from(node: Element, graph):
    ''' returns an iterator over the (node, edge)s that can be reached from
    node '''
    for edge in graph.findall(".//*[@from='{}']".format(node.attrib['id'])):
        yield (graph.find(".//*[@id='{}']".format(edge.attrib['to'])), edge)


@comment
def has_no_incoming(node: Element, graph: 'root Element'):
    ''' returns true if this node has no edges pointing to it '''
    return len(graph.findall(".//*[@to='{}']".format(node.attrib['id']))) == 0


@comment
def has_edges(graph: 'root Element'):
    ''' returns true if the graph still has edge elements '''
    return len(graph.findall("./connector")) > 0


@comment
def topological_sort(start_node: Element, graph: 'Element') -> 'ElementTree':
    ''' sorts topologically the given xml element tree, source:
    https://en.wikipedia.org/wiki/Topological_sorting '''
    sorted_elements = []  # Empty list that will contain the sorted elements
    no_incoming = [(start_node, None)]  # (node, edge that points to this node)

    while len(no_incoming) > 0:
        node, edge = no_incoming.pop()

        if edge is not None:
            sorted_elements.append(edge)
        sorted_elements.append(node)

        for m, edge in nodes_from(node, graph=graph):
            graph.remove(edge)

            if has_no_incoming(m, graph):
                no_incoming.append((m, edge))

    if has_edges(graph) > 0:
        raise Exception('graph is cyclic')

    return etree_from_list(graph, sorted_elements)


SUPPORTED_ATTRS = {
    'type': str,
    'name': str,
    'required': lambda x: bool(x),
    'regex': str,
    'label': str,
    'placeholder': str,
    'default': str,
    'helper': str,
}


def form_to_dict(form):
    inputs = form.getElementsByTagName('input')

    form_dict = {
        'ref': get_ref(form),
        'inputs': [],
    }

    for input in inputs:
        input_attrs = [
            (attr, func(input.getAttribute(attr)))
            for attr, func in SUPPORTED_ATTRS.items()
        ] + [('options', list(map(
            lambda e: {
                'value': e.getAttribute('value'),
                'label': e.firstChild.nodeValue,
            },
            input.getElementsByTagName('option'),
        )))]

        form_dict['inputs'].append(dict(filter(
            lambda a: a[1],
            input_attrs
        )))

    return form_dict



from ldap3 import Server, Connection, ALL, NTLM, core
from ldap3.core.exceptions import LDAPBindError, LDAPSocketOpenError

from cacahuate.auth.base import BaseAuthProvider, BaseUser
from cacahuate.errors import AuthenticationError
from cacahuate.http.wsgi import app


class LdapUser(BaseUser):

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def get_identifier(self):
        return self.username


class LdapAuthProvider(BaseAuthProvider):

    def authenticate(self, **credentials):
        if 'username' not in credentials or \
           'password' not in credentials:
            raise AuthenticationError

        domain = app.config['LDAP_DOMAIN']
        if 'domain' in credentials:
            domain = credentials['domain']

        username = '{domain}\\{username}'.format(
            domain=domain,
            username=credentials['username'],
        )

        password = credentials['password']

        server = Server(
            app.config['LDAP_URI'],
            get_info=ALL,
            use_ssl=app.config['LDAP_SSL'],
        )

        try:
            conn = Connection(
                server,
                user=username,
                password=password,
                auto_bind=True,
                authentication=NTLM,
            )
        except LDAPBindError:
            raise AuthenticationError

        return LdapUser(
            username=username
        )



from importlib import import_module
from xml.dom.minidom import Element
from case_conversion import pascalcase
from flask import request

from cacahuate.errors import ValidationErrors, InputError,\
    RequiredInputError, HierarchyError
from cacahuate.http.errors import BadRequest, Unauthorized, Forbidden
from cacahuate.models import User, Token
from cacahuate.xml import get_ref, resolve_params
from cacahuate.http.wsgi import app


def get_associated_data(ref: str, data: dict) -> dict:
    ''' given a reference returns its asociated data in the data dictionary '''
    if 'form_array' not in data:
        return {}

    for form in data['form_array']:
        if type(form) != dict:
            continue

        if 'ref' not in form:
            continue

        if form['ref'] == ref:
            return form['data']

    return {}


def validate_input(form_index: int, input: Element, value):
    ''' Validates the given value against the requirements specified by the
    input element '''
    if input.getAttribute('required') and (value == '' or value is None):
        raise RequiredInputError(form_index, input.getAttribute('name'))

    return value


def validate_form(index: int, form: Element, data: dict) -> dict:
    ''' Validates the given data against the spec contained in form. In case of
    failure raises an exception. In case of success returns the validated data.
    '''
    ref = get_ref(form)

    given_data = get_associated_data(ref, data)
    collected_data = {}
    errors = []

    for input in form.getElementsByTagName('input'):
        name = input.getAttribute('name')

        try:
            collected_data[name] = \
                validate_input(index, input, given_data.get(name))
        except InputError as e:
            errors.append(e)

    if errors:
        raise ValidationErrors(errors)

    return collected_data


def validate_json(json_data: dict, req: list):
    errors = []

    for item in req:
        if item not in json_data:
            errors.append({
                'detail': '{} is required'.format(item),
                'where': 'request.body.{}'.format(item),
                'code': 'validation.required',
            })

    if errors:
        raise BadRequest(errors)


def validate_auth(node, execution=None):
    auth = node.getElementsByTagName('auth')

    if len(auth) == 0:
        return None, None

    auth_node = auth[0]

    # Authorization required but not provided, notify
    if request.authorization is None:
        raise Unauthorized([{
            'detail': 'You must provide basic authorization headers',
            'where': 'request.authorization',
        }])

    identifier = request.authorization['username']
    token = request.authorization['password']

    user = User.get_by('identifier', identifier)
    token = Token.get_by('token', token)

    if user is None or token is None or token.proxy.user.get().id != user.id:
        raise Unauthorized([{
            'detail': 'Your credentials are invalid, sorry',
            'where': 'request.authorization',
        }])

    # check for filters for this user
    filter_q = auth_node.getElementsByTagName('filter')

    if len(filter_q) == 1:
        filter_node = filter_q[0]
        backend = filter_node.getAttribute('backend')

        mod = import_module('cacahuate.auth.hierarchy.{}'.format(backend))
        HiPro = getattr(mod, pascalcase(backend) + 'HierarchyProvider')

        hipro = HiPro(app.config)

        try:
            hipro.validate_user(user, **resolve_params(filter_node, execution))
        except HierarchyError:
            raise Forbidden([{
                'detail': 'The provided credentials do not match the specified'
                          ' hierarchy',
                'where': 'request.authorization',
            }])

    return get_ref(auth_node), user


def validate_forms(node):
    form_array = node.getElementsByTagName('form-array')
    collected_forms = []

    if len(form_array) == 0:
        return []

    form_array_node = form_array[0]

    errors = []

    for index, form in enumerate(form_array_node.getElementsByTagName('form')):
        try:
            data = validate_form(index, form, request.json)
            collected_forms.append((get_ref(form), data))
        except ValidationErrors as e:
            errors += e.errors

    if len(errors) > 0:
        raise BadRequest(ValidationErrors(errors).to_json())

    return collected_forms



from coralillo.errors import ValidationErrors, ModelNotFoundError, BadField
from flask import jsonify

from cacahuate.http.wsgi import app
from cacahuate.http.errors import JsonReportedException


@app.errorhandler(JsonReportedException)
def json_formatted_handler(e):
    return jsonify(e.to_json()), e.status_code, e.headers


@app.errorhandler(404)
def handle_404(e):
    return jsonify({
        'errors': [{
            'detail': e.description,
            'where': 'request.url',
        }],
    }), e.code


@app.errorhandler(401)
def handle_401(e):
    return jsonify({
        'errors': [{
            'detail': ':'.join(str(e).split(':')[1:])[1:],
            'where': 'request.body',
        }],
    }), e.code


@app.errorhandler(500)
def handle_500(e):
    return jsonify({
        'errors': [{
            'detail': 'The server has failed its mission',
            'where': 'server',
        }],
    }), 500



from cacahuate.auth.base import BaseAuthProvider, BaseUser
from cacahuate.errors import AuthenticationError


class HardcodedUser(BaseUser):

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def get_identifier(self):
        return self.username


class HardcodedAuthProvider(BaseAuthProvider):

    def authenticate(self, **credentials):
        if 'username' not in credentials or \
           'password' not in credentials:
            raise AuthenticationError

        username = credentials['username']
        password = credentials['password']

        if username != 'juan' or password != '123456':
            raise AuthenticationError

        return HardcodedUser(
            username=username,
            password=password,
        )



from lark import Lark, Transformer
import operator
import os

from cacahuate.errors import RefNotFound


class Condition:

    def __init__(self, execution):
        filename = os.path.join(
                                os.path.dirname(__file__),
                                'grammars/condition.g'
                                )

        with open(filename) as grammar_file:
            self.parser = Lark(
                grammar_file.read(),
                start='condition',
                parser='lalr',
                transformer=self.ConditionTransformer(execution),
            )

    def parse(self, string):
        return self.parser.parse(string)

    class ConditionTransformer(Transformer):

        def __init__(self, execution):
            self._execution = execution

        def op_eq(self, _):
            return operator.eq

        def op_ne(self, _):
            return operator.ne

        def type_form(self, _):
            return 'forms'

        def variable(self, args):
            return args[0][:]

        def obj_id(self, args):
            return args[0][:]

        def member(self, args):
            return args[0][:]

        def ref(self, args):
            obj_type, obj_id, member = args

            try:
                obj = next(
                        getattr(
                                self._execution.proxy,
                                obj_type).q().filter(ref='#'+obj_id)
                        )
            except StopIteration:
                raise RefNotFound

            return obj.data.get(member)

        def string(self, args):
            return args[0][1:-1]

        def condition(self, args):
            left, op, right = args

            return op(left, right)



from flask import abort
from importlib import import_module
from cacahuate.http.errors import NotFound
from cacahuate.http.wsgi import app
from werkzeug.routing import BaseConverter
import case_conversion
import os
import sys


class AuthProviderConverter(BaseConverter):

    def to_python(self, value):
        # this allows custom login providers
        if value in app.config['LOGIN_PROVIDERS']:
            import_path = app.config['LOGIN_PROVIDERS'][value]

            cwd = os.getcwd()

            if cwd not in sys.path:
                sys.path.insert(0, cwd)
        else:
            import_path = 'cacahuate.auth.backends.' + value

        try:
            mod = import_module(import_path)
            cls = getattr(
                mod, case_conversion.pascalcase(value) + 'AuthProvider'
            )
        except ModuleNotFoundError:
            abort(404, 'Auth backend not found: {}'.format(value))
        except AttributeError as e:
            abort(500, 'Misconfigured auth provider, sorry')

        return cls(app.config)

    def to_url(self, values):
        raise NotImplementedError('this converter does not work backwards')


app.url_map.converters['AuthProvider'] = AuthProviderConverter



""" Here is defined the node class and its subclasses, which define the kinds
of directions that this virtual machine can follow """
import case_conversion
from typing import Iterator
from xml.dom.minidom import Element

from cacahuate.xml import Xml
from cacahuate.logger import log
from cacahuate.grammar import Condition
from cacahuate.errors import ElementNotFound, IncompleteBranch


class Node:
    ''' A node from the process's graph. It is initialized from an Element
    '''

    def __init__(self, element):
        self.element = element

    def next(self, xmliter: Iterator[Element], execution) -> ['Node']:
        ''' Gets the next node in the graph,
        if it fails raises an exception.'''
        raise NotImplementedError('Should be implemented for subclasses')


class EndNode(Node):

    def next(self, xml, execution):
        return []


class SimpleNode(Node):

    def next(self, xml: Xml, execution) -> ['Node']:
        ''' just find the next node in the graph '''
        def find_node(e):
            if e.tagName != 'connector':
                return False

            return e.getAttribute('from') == self.element.getAttribute('id')

        conn = xml.find(find_node)

        return [make_node(xml.find(
            lambda e: e.getAttribute('id') == conn.getAttribute('to')
        ))]


class DecisionNode(Node):

    def next(self, xml: Xml, execution) -> ['Node']:
        ''' find node whose value corresponds to the answer '''
        def find_node(el):
            if el.tagName != 'connector':
                return False

            if el.getAttribute('from') != self.element.getAttribute('id'):
                return False

            cons = el.getElementsByTagName('condition')

            if len(cons) != 1:
                return False

            con = cons[0]
            con.normalize()

            return Condition(execution).parse(con.firstChild.nodeValue)

        try:
            conn = xml.find(find_node)
        except ElementNotFound:
            raise IncompleteBranch(
                'Either not all branches for this desition are defined or '
                'there is not enough information in the asociated data'
            )

        return [make_node(xml.find(
            lambda e: e.getAttribute('id') == conn.getAttribute('to')
        ))]


def make_node(element):
    ''' returns a build Node object given an Element object '''
    if not element.getAttribute('class'):
        raise KeyError('Must have the class atrribute')

    class_name = case_conversion.pascalcase(
                                            element.getAttribute('class')
                ) + 'Node'
    available_classes = __import__(__name__).node

    if class_name not in dir(available_classes):
        raise ValueError('Class definition not found: {}'.format(class_name))

    return getattr(available_classes, class_name)(
        element
    )



from cacahuate.auth.base import BaseHierarchyProvider
from cacahuate.auth.backends.hardcoded import HardcodedUser
from cacahuate.errors import HierarchyError
from cacahuate.models import User


class HardcodedHierarchyProvider(BaseHierarchyProvider):

    def validate_user(self, user, **params):
        base_user = User.get_by('identifier', params.get('employee'))

        if base_user is None:
            raise HierarchyError

        relation = params.get('relation')
        manager_identifier = '{}_{}'.format(base_user.identifier, relation)
        manager = User.get_by('identifier', manager_identifier)

        if manager is None or manager.id != user.id:
            raise HierarchyError

    def find_users(self, **params):
        employee = params.get('employee')
        relation = params.get('relation')

        return list(map(
            lambda u: HardcodedUser(username=u.identifier),
            User.q().filter(
                identifier='{}_{}'.format(employee, relation)
            )
        ))



from coralillo.errors import ModelNotFoundError
from datetime import datetime
from flask import g
from flask import request, jsonify, json
import os
import pika

from cacahuate.errors import ProcessNotFound, ElementNotFound, MalformedProcess
from cacahuate.http.errors import BadRequest, NotFound, UnprocessableEntity, \
    Forbidden
from cacahuate.http.forms import ContinueProcess
from cacahuate.http.middleware import requires_json, requires_auth
from cacahuate.http.validation import validate_forms, validate_json, \
    validate_auth
from cacahuate.http.wsgi import app, mongo
from cacahuate.models import Execution, Pointer, User, Token, Activity, \
    Questionaire
from cacahuate.rabbit import get_channel
from cacahuate.xml import Xml, form_to_dict


def trans_id(obj):
    obj['_id'] = str(obj['_id'])
    return obj


def trans_date(obj):
    obj['started_at'] = \
        obj['started_at'].isoformat() if (
                                        obj['started_at'] is not None
                                        ) else None
    obj['finished_at'] = \
        obj['finished_at'].isoformat() if (
                                        obj['finished_at'] is not None
                                        ) else None
    return obj


@app.route('/', methods=['GET', 'POST'])
@requires_json
def index():
    if request.method == 'GET':
        return {
            'hello': 'world',
        }
    elif request.method == 'POST':
        return request.json


@app.route('/v1/execution', methods=['POST'])
@requires_json
def start_process():
    validate_json(request.json, ['process_name'])

    try:
        xml = Xml.load(app.config, request.json['process_name'])
    except ProcessNotFound as e:
        raise NotFound([{
            'detail': '{} process does not exist'
                      .format(request.json['process_name']),
            'where': 'request.body.process_name',
        }])
    except MalformedProcess as e:
        raise UnprocessableEntity([{
            'detail': '{} process lacks important nodes and structure'
                      .format(request.json['process_name']),
            'where': 'request.body.process_name',
        }])

    try:
        start_point = xml.start_node()
    except ElementNotFound as e:
        raise UnprocessableEntity([{
            'detail': '{} process does not have a start node, thus cannot be '
                      'started'.format(request.json['process_name']),
            'where': 'request.body.process_name',
        }])

    # Check for authorization
    auth_ref, user = validate_auth(start_point)

    # check if there are any forms present
    collected_forms = validate_forms(start_point)

    # save the data
    execution = Execution(
        process_name=xml.filename,
    ).save()

    pointer = Pointer(
        node_id=start_point.getAttribute('id'),
    ).save()

    pointer.proxy.execution.set(execution)

    actors = []
    if auth_ref is not None:
        activity = Activity(ref=auth_ref).save()
        activity.proxy.user.set(user)
        activity.proxy.execution.set(execution)
        actors.append({'ref': auth_ref, 'user': user.to_json()})
    forms = []

    if len(collected_forms) > 0:
        for ref, form_data in collected_forms:
            ques = Questionaire(ref=ref, data=form_data).save()
            ques.proxy.execution.set(execution)
            forms.append({'ref': ref, 'data': form_data})

    # log to mongo
    collection = mongo.db[app.config['MONGO_HISTORY_COLLECTION']]

    collection.insert_one({
        'started_at': datetime.now(),
        'finished_at': datetime.now(),
        'execution_id': execution.id,
        'node_id': start_point.getAttribute('id'),
        'forms': forms,
        'actors': actors,
        'documents': []
    })

    # trigger rabbit
    channel = get_channel()
    channel.basic_publish(
        exchange='',
        routing_key=app.config['RABBIT_QUEUE'],
        body=json.dumps({
            'command': 'step',
            'process': execution.process_name,
            'pointer_id': pointer.id,
        }),
        properties=pika.BasicProperties(
            delivery_mode=2,
        ),
    )

    return {
        'data': execution.to_json(),
    }, 201


@app.route('/v1/pointer', methods=['POST'])
@requires_json
def continue_process():
    validate_json(request.json, ['execution_id', 'node_id'])

    execution_id = request.json['execution_id']
    node_id = request.json['node_id']

    try:
        execution = Execution.get_or_exception(execution_id)
    except ModelNotFoundError:
        raise BadRequest([{
            'detail': 'execution_id is not valid',
            'code': 'validation.invalid',
            'where': 'request.body.execution_id',
        }])

    xml = Xml.load(app.config, execution.process_name)

    try:
        continue_point = xml.find(lambda e: e.getAttribute('id') == node_id)
    except ElementNotFound as e:
        raise BadRequest([{
            'detail': 'node_id is not a valid node',
            'code': 'validation.invalid_node',
            'where': 'request.body.node_id',
        }])

    try:
        pointer = next(execution.proxy.pointers.q().filter(node_id=node_id))
    except StopIteration:
        raise BadRequest([{
            'detail': 'node_id does not have a live pointer',
            'code': 'validation.no_live_pointer',
            'where': 'request.body.node_id',
        }])

    # Check for authorization
    auth_ref, user = validate_auth(continue_point, execution)

    # Validate asociated forms
    collected_forms = validate_forms(continue_point)

    # save the data
    actors = []

    if auth_ref is not None:
        activity = Activity(ref=auth_ref).save()
        activity.proxy.user.set(user)
        activity.proxy.execution.set(execution)

        actors.append({'ref': auth_ref, 'user': user.to_json()})

    forms = []
    if len(collected_forms) > 0:
        for ref, form_data in collected_forms:
            ques = Questionaire(ref=ref, data=form_data).save()
            ques.proxy.execution.set(execution)
            forms.append({'ref': ref, 'data': form_data})

    # trigger rabbit
    channel = get_channel()
    channel.basic_publish(
        exchange='',
        routing_key=app.config['RABBIT_QUEUE'],
        body=json.dumps({
            'command': 'step',
            'process': execution.process_name,
            'pointer_id': pointer.id,
            'forms': forms,
            'actors':  actors,
            'documents': []
        }),
        properties=pika.BasicProperties(
            delivery_mode=2,
        ),
    )

    return {
        'data': 'accepted',
    }, 202


@app.route('/v1/process', methods=['GET'])
def list_process():
    def add_form(xml):
        try:
            start_node = xml.start_node()
        except ElementNotFound:
            return None

        json_xml = xml.to_json()
        forms = []

        for form in start_node.getElementsByTagName('form'):
            forms.append(form_to_dict(form))

        json_xml['form_array'] = forms

        return json_xml

    return jsonify({
        'data': list(filter(
            lambda x: x,
            map(
                add_form,
                Xml.list(app.config),
            )
        ))
    })


@app.route('/v1/activity', methods=['GET'])
@requires_auth
def list_activities():
    activities = g.user.proxy.activities.get()

    return jsonify({
        'data': list(map(
            lambda a: a.to_json(embed=['execution']),
            activities
        )),
    })


@app.route('/v1/activity/<id>', methods=['GET'])
@requires_auth
def one_activity(id):
    try:
        activity = Activity.get_or_exception(id)
    except ModelNotFoundError:
        raise BadRequest([{
            'detail': 'activity_id is not valid',
            'code': 'validation.invalid',
            'where': 'request.body.execution_id',
        }])

    user_activity = User.get_or_exception(activity.user)
    if not g.user == user_activity:
        raise Forbidden([{
            'detail': 'You must provide basic authorization headers',
            'where': 'request.authorization',
        }])

    return jsonify({
        'data': activity.to_json(),
    })


@app.route('/v1/log/<id>', methods=['GET'])
def list_logs(id):
    collection = mongo.db[app.config['MONGO_HISTORY_COLLECTION']]
    node_id = request.args.get('node_id')
    query = {'execution_id': id}

    if node_id:
        query['node_id'] = node_id

    return jsonify({
        "data": list(map(
            trans_date,
            map(
                trans_id,
                collection.find(query)
            )
        )),
    }), 200



from case_conversion import pascalcase
from coralillo.errors import ModelNotFoundError
from datetime import datetime
from importlib import import_module
from pymongo import MongoClient
import simplejson as json
import pika

from cacahuate.errors import CannotMove
from cacahuate.logger import log
from cacahuate.models import Execution, Pointer
from cacahuate.node import make_node, Node
from cacahuate.xml import Xml, resolve_params
from cacahuate.auth.base import BaseUser


class Handler:
    ''' The actual process machine, it is in charge of moving the pointers
    among the graph of nodes '''

    def __init__(self, config):
        self.config = config
        self.mongo = None

    def __call__(self, channel, method, properties, body: bytes):
        ''' the main callback of cacahuate '''
        message = self.parse_message(body)

        try:
            to_queue = self.call(message, channel)
        except ModelNotFoundError as e:
            return log.error(str(e))
        except CannotMove as e:
            return log.error(str(e))

        for pointer in to_queue:
            channel.basic_publish(
                exchange='',
                routing_key=self.config['RABBIT_QUEUE'],
                body=json.dumps({
                    'command': 'step',
                    'pointer_id': pointer.id,
                }),
                properties=pika.BasicProperties(
                    delivery_mode=2,
                ),
            )

        if not self.config['RABBIT_NO_ACK']:
            channel.basic_ack(delivery_tag=method.delivery_tag)

    def call(self, message: dict, channel):
        execution, pointer, xml, cur_node, *rest = self.recover_step(message)

        to_queue = []  # pointers to be created

        # node's lifetime ends here
        self.teardown(pointer, *rest)
        next_nodes = cur_node.next(xml, execution)

        for node in next_nodes:
            # node's begining of life
            pointer = self.wakeup(node, execution, channel)

            # async nodes don't return theirs pointers so they are not queued
            if pointer:
                to_queue.append(pointer)

        if execution.proxy.pointers.count() == 0:
            execution.delete()

        return to_queue

    def wakeup(self, node, execution, channel):
        ''' Waking up a node often means to notify someone or something about
        the execution, this is the first step in node's lifecycle '''
        # create a pointer in this node
        pointer = self.create_pointer(node, execution)

        # notify someone
        filter_q = node.element.getElementsByTagName('filter')

        if len(filter_q) == 0:
            return

        filter_node = filter_q[0]
        backend = filter_node.getAttribute('backend')

        mod = import_module('cacahuate.auth.hierarchy.{}'.format(backend))
        HierarchyProvider = getattr(
            mod,
            pascalcase(backend) + 'HierarchyProvider'
        )

        hierarchy_provider = HierarchyProvider(self.config)
        husers = hierarchy_provider.find_users(
            **resolve_params(filter_node, execution)
        )

        for huser in husers:
            user = huser.get_user()

            if pointer:
                user.proxy.tasks.add(pointer)

            mediums = self.get_contact_channels(user)

            for medium, params in mediums:
                channel.basic_publish(
                    exchange=self.config['RABBIT_NOTIFY_EXCHANGE'],
                    routing_key=medium,
                    body=json.dumps(params),
                    properties=pika.BasicProperties(
                        delivery_mode=2,
                    ),
                )

        # update registry about this pointer
        collection = self.get_mongo()

        collection.insert_one({
            'started_at': datetime.now(),
            'finished_at': None,
            'execution_id': execution.id,
            'node_id': node.element.getAttribute('id'),
            'forms': [],
            'documents': [],
            'actors': [],
        })

        # nodes with forms and documents are not queued
        if len(node.element.getElementsByTagName('form-array')) > 0:
            return pointer

    def teardown(self, pointer, forms, actors, documents):
        ''' finishes the node's lifecycle '''
        collection = self.get_mongo()

        collection.update_one({
            'execution_id': pointer.proxy.execution.get().id,
            'node_id': pointer.node_id,
        }, {
            '$set': {
                'finished_at': datetime.now(),
                'forms': forms,
                'actors': actors,
                'documents': documents,
            },
        })

        pointer.delete()

    def parse_message(self, body: bytes):
        ''' validates a received message against all possible needed fields
        and structure '''
        try:
            message = json.loads(body)
        except json.decoder.JSONDecodeError:
            raise ValueError('Message is not json')

        if 'command' not in message:
            raise KeyError('Malformed message: must contain command keyword')

        if message['command'] not in self.config['COMMANDS']:
            raise ValueError('Command not supported: {}'.format(
                message['command']
            ))

        return message

    def get_mongo(self):
        if self.mongo is None:
            client = MongoClient()
            db = client[self.config['MONGO_DBNAME']]

            self.mongo = db[self.config['MONGO_HISTORY_COLLECTION']]

        return self.mongo

    def get_contact_channels(self, user: BaseUser):
        return [('email', {})]

    def create_pointer(self, node: Node, execution: Execution):
        ''' Given a node, its process, and a specific execution of the former
        create a persistent pointer to the current execution state '''
        pointer = Pointer \
            .validate(node_id=node.element.getAttribute('id')) \
            .save()
        pointer.proxy.execution.set(execution)

        return pointer

    def recover_step(self, message: dict):
        ''' given an execution id and a pointer from the persistent storage,
        return the asociated process node to continue its execution '''
        if 'pointer_id' not in message:
            raise KeyError('Requested step without pointer id')

        pointer = Pointer.get_or_exception(message['pointer_id'])
        execution = pointer.proxy.execution.get()
        xml = Xml.load(self.config, execution.process_name)

        assert execution.process_name == xml.filename, 'Inconsistent pointer'

        point = xml.find(
            lambda e: e.getAttribute('id') == pointer.node_id
        )

        return (
            execution,
            pointer,
            xml,
            make_node(point),
            message.get('forms', []),
            message.get('actors', []),
            message.get('documents', []),
        )



from coralillo.errors import BadField


class AuthenticationError(Exception):

    pass


class ProcessNotFound(Exception):

    pass


class ElementNotFound(Exception):

    pass


class CannotMove(Exception):

    pass


class RefNotFound(Exception):

    pass


class MalformedProcess(Exception):

    pass


class HierarchyError(Exception):

    pass


class IncompleteBranch(Exception):

    pass


class NoPointerAlive(BadField):

    message = '{field} does not have a live pointer'
    errorcode = 'no_live_pointer'


class InputError(Exception):

    detail = None
    code = None
    form_index = 0

    def __init__(self, form_index, input):
        self.input = input
        self.form_index = form_index

    def to_json(self):
        return {
            'detail': self.detail.format(input=self.input),
            'where': 'request.body.form_array.{}.{}'.format(
                self.form_index,
                self.input,
            ),
            'code': self.code,
        }


class RequiredInputError(InputError):

    detail = "'{input}' input is required"
    code = 'validation.required'


class ValidationErrors(Exception):

    def __init__(self, errors):
        super().__init__()
        self.errors = errors

    def to_json(self):
        return list(map(
            lambda e: e.to_json(),
            self.errors
        ))



from coralillo import Form, fields, errors
from coralillo.validation import validation_rule

from cacahuate.errors import ElementNotFound, NoPointerAlive


class ContinueProcess(Form):
    execution_id = fields.Text()
    node_id = fields.Text()

    @validation_rule
    def custom_validations(obj):
        ''' Ensures that execution_id points to a currently running execution,
        that node_id is a real node from the same process file as the provided
        execution and that a pointer pointing to that node belonging to the
        given execution exists '''
        # validates the existence of the execution
        if not obj.execution_id or not obj.node_id:
            return  # previous validation didn't pass

        from cacahuate.models import Execution

        execution = Execution.get(obj.execution_id)

        if execution is None:
            raise errors.InvalidFieldError(field='execution_id')

        # validates the existence of the node
        from cacahuate.xml import Xml
        from cacahuate.http.wsgi import app

        xml = Xml.load(app.config, execution.process_name)

        def testfunc(e):
            return e.getAttribute('id') == obj.node_id

        try:
            xml.find(testfunc)
        except ElementNotFound:
            raise errors.InvalidFieldError(field='node_id')

        # validates the existence of a pointer asociated
        # to the node and execution
        ptrs = execution.proxy.pointers.get()
        pointer = None

        for ptr in ptrs:
            if ptr.node_id == obj.node_id:
                pointer = ptr
                break
        else:
            raise NoPointerAlive(field='node_id')

        obj.execution = execution
        obj.pointer = pointer


def bind_forms(engine):
    ContinueProcess.set_engine(engine)
